import json
import codecs
import uuid
from pathlib import Path
import shutil
from urllib.parse import urlparse
from .exceptions import DataNotExistsError, DataIsNotDirError, DataInfoNotFoundError
from .data_file import DataFile


class Data:
    """
    Test data

    
    """

    info_filename = '.lyrebird_prop'

    def __init__(self, data_id, data_name, data_path, data_rule=None):
        self.id = data_id
        self.name = data_name
        self.path = data_path
        self.rule = data_rule

        self.request = DataFile(self, 'request', filetype='json')
        self.request_data = DataFile(self, 'request_data')
        self.response = DataFile(self, 'response', filetype='json')
        self.response_data = DataFile(self, 'response_data')


    @classmethod
    def createify(cls, data_path):
        new_path = Path(data_path)
        if not new_path.exists():
            raise DataNotExistsError(data_path)
        if not new_path.is_dir():
            raise DataIsNotDirError(data_path)
        # Read data info
        with codecs.open(new_path/cls.info_filename, 'r' , 'utf-8') as f:
            info = json.load(f)
        if not info:
            raise DataInfoNotFoundError(new_path)
        
        _id = info.get('id')
        _name = info.get('name')
        _rule = info.get('rule')
        return cls(_id, _name, new_path, _rule)

    @classmethod
    def new_data(cls, data_dir, flow=None, request=None, request_data=None, response=None, response_data=None):
        _id = str(uuid.uuid4())
        _name = _id

        _url = None
        if flow:
            _url = flow.get('request', {'url': None}).get('url')
        elif request:
            _req = json.loads(request)
            _url = request.get('url')
        
        _rule = {'request.url':None}

        if _url:
            parsed_url = urlparse(_url)
            if parsed_url.path is not '':
                _name = parsed_url.path
            else:
                _name = parsed_url.hostname
            _rule['request.url'] = _name

        _data_path = Path(data_dir)/_id
        _data_path.mkdir(parents=True, exist_ok=True)

        data = cls(_id, _name, _data_path)
        
        data.rule = _rule

        if flow:
            data.set_flow(flow)
        else:
            data.request.content = request
            data.request_data.content = request_data
            data.response.content = response
            data.response_data.content = response_data
        return data

    def set_flow(self, flow):
        if flow.get('request'):
            request_obj = {
                'method': flow['request'].get('method'),
                'url': flow['request'].get('url'),
                'headers': flow['request'].get('headers')
            }
            content_type = request_obj['headers'].get('Content-Type')
            if content_type and 'json' in content_type:
                self.request_data.filetype = 'json'
            elif content_type and 'html' in content_type:
                self.request_data.filetype = 'html'
            elif content_type and 'xml' in content_type:
                self.request_data.filetype = 'xml'
            elif content_type and 'text' in content_type:
                self.request_data.filetype = 'text'

            self.request.content = json.dumps(request_obj, ensure_ascii=False, indent=4)
            if flow['request'].get('data'):
                self.request_data.content = flow['request'].get('data')

        if flow.get('response'):
            response_obj = {
                'code': flow['response'].get('code'),
                'headers': flow['response'].get('headers')
            }
            content_type = response_obj['headers'].get('Content-Type')
            if content_type and 'json' in content_type:
                self.response_data.filetype = 'json'
            elif content_type and 'html' in content_type:
                self.response_data.filetype = 'html'
            elif content_type and 'xml' in content_type:
                self.response_data.filetype = 'xml'
            elif content_type and 'text' in content_type:
                self.response_data.filetype = 'text'
            
            self.response.content = json.dumps(response_obj, ensure_ascii=False, indent=4)
            if flow['response'].get('data'):
                if self.response_data.filetype == 'json':
                    self.response_data.content = json.dumps(flow['response'].get('data'))
                else:
                    self.response_data.content = flow['response'].get('data')


    def save(self):
        with codecs.open(self.path/self.info_filename, 'w', 'utf-8') as f:
            info = {
                'id': self.id,
                'name': self.name,
                'rule': self.rule,
                'data_files': {
                    'request': {
                        'filetype': self.request.filetype
                    },
                    'request_data': {
                        'filetype': self.request_data.filetype
                    },
                    'response':{
                        'filetype': self.response.filetype
                    },
                    'response_data': {
                        'filetype': self.response_data.filetype
                    }
                }
            }
            if self.request.content:
                info['url'] = json.loads(self.request.content).get('url')
            json.dump(info, f, ensure_ascii=False, indent=4)

    
    def delete(self):
        shutil.rmtree(self.path)

    def _get_content_type_from_content(self, content):
        filetype = 'bin'
        content_obj = json.loads(content)
        content_type = content_obj.get('headers', {}).get('Content-Type')
        if content_type and 'json' in content_type:
            filetype = 'json'
        elif content_type and 'html' in content_type:
            filetype = 'html'
        elif content_type and 'xml' in content_type:
            filetype = 'xml'
        elif content_type and 'text' in content_type:
            filetype = 'text'
        return filetype

    def json(self, detail=False):
        json_obj = {
            'id': self.id,
            'name': self.name
        }
        if detail:
            json_obj['rule'] = self.rule
            if self.request.content:
                json_obj['request'] = self.request.json()
                self.request_data.filetype = self._get_content_type_from_content(self.request.content)
            if self.request_data.content:
                json_obj['request_data'] = self.request_data.json()
            if self.response.content:
                json_obj['response'] = self.response.json()
            if self.response_data.content:
                self.response_data.filetype = self._get_content_type_from_content(self.response.content)
                json_obj['response_data'] = self.response_data.json()
        return json_obj



import re


class MockRouter:
    def __init__(self):
        self.data_map = []

    def switch_group(self, data_group):
        self.data_map = []
        if not data_group:
            return
        for data_id in data_group.all_data:
            data = data_group.all_data[data_id]
            self.data_map.append((MockRule(data.rule), data))

    def get_mock_data(self, flow):
        for rule, data in self.data_map:
            if rule.match(flow):
                return data
        # TODO none return

class MockRule:
    def __init__(self, rule):
        self.rule = rule

    def match(self, flow):
        if not self.rule:
            return False
        for rule_name in self.rule:
            pattern = self.rule[rule_name]
            target = self._get_target_content(rule_name, flow)
            if not re.search(pattern, target):            
                return False
        return True

    def _get_target_content(self, rule_name, flow):
        prop_keys = rule_name.split('.')
        result = flow
        for prop_key in prop_keys:
            result = result.get(prop_key)
            if not result:
                return None
        return result



import logging
from colorama import Fore, Style, Back
from collections import namedtuple
from lyrebird import application


_stream_logger_inited = False


def get_logger()->logging.Logger:
    global _stream_logger_inited
    if not _stream_logger_inited:
        _init_stream_logger()
        _stream_logger_inited = True
    return logging.getLogger('lyrebird')


Color = namedtuple('Color', ['fore', 'style', 'back'])

COLORS = dict(
    CRITICAL=Color(fore=Fore.WHITE, style=Style.BRIGHT, back=Back.RED),
    ERROR=Color(fore=Fore.RED, style=Style.NORMAL, back=Back.RESET),
    WARNING=Color(fore=Fore.YELLOW, style=Style.NORMAL, back=Back.RESET),
    INFO=Color(fore=Fore.WHITE, style=Style.NORMAL, back=Back.RESET),
    DEBUG=Color(fore=Fore.GREEN, style=Style.NORMAL, back=Back.RESET)
)

def colorit(message, levelname):
    color = COLORS.get(levelname)
    if color:
        return f'{color.fore}{color.style}{color.back}{message}{Style.RESET_ALL}'
    else:
        return message


class ColorFormater(logging.Formatter):

    def format(self, record:logging.LogRecord):
        module = f'{colorit(record.module, record.levelname)}'
        msg = f'{colorit(record.msg, record.levelname)}'
        levelname = f'{colorit(record.levelname, record.levelname)}'
        return f'{levelname} [{module}] {msg}'


def _init_stream_logger():
    logger:logging.Logger = logging.getLogger('lyrebird')
    
    color_formater = ColorFormater(fmt='%(levelname)s [%(module)s] %(message)s')
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(color_formater)
    logger.addHandler(stream_handler)


def _init_file_logger(custom_log_path=None):
    file_formater = logging.Formatter(
        fmt='%(asctime)s %(levelname)s [%(module)s] - %(threadName)s [PID] %(process)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    if custom_log_path:
        log_file = custom_log_path
    else:
        log_file = application.root_dir()/'lyrebird.log'

    file_handler = logging.handlers.TimedRotatingFileHandler(log_file, backupCount=1, encoding='utf-8', when='midnight')
    file_handler.setFormatter(file_formater)
    
    logger:logging.Logger = logging.getLogger('lyrebird')
    logger.addHandler(file_handler)


def _setup_3rd_loggers():
    socketio = logging.getLogger('socketio')
    socketio.setLevel(logging.WARNING)

    engineio = logging.getLogger('engineio')
    engineio.setLevel(logging.WARNING)

    mock = logging.getLogger('mock')
    mock.setLevel(logging.WARNING)
    mock.addHandler(logging.StreamHandler())

    werkzeug = logging.getLogger('werkzeug')
    werkzeug.setLevel(logging.WARNING)

def init(custom_log_path=None):    
    _init_file_logger(custom_log_path=custom_log_path)
    _setup_3rd_loggers()

    logger:logging.Logger = logging.getLogger('lyrebird')
    if application.config.get('verbose', False):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)


import json
import datetime
import lyrebird
from lyrebird.mock import context
from sqlalchemy import Column, Table, String, Integer, DateTime, Text


class Flow(context.db.model):

    id = Column(Integer, primary_key=True, autoincrement=True)
    channel = Column(String(16))
    content = Column(Text)
    ts = Column(DateTime(timezone=True), default=datetime.datetime.now)

    def __init__(self, channel, content):
        self.channel = channel
        self.content = content

def active_db_listener():
    lyrebird.subscribe('any', save_data_into_db, event=True)
    context.db.create_all()

def save_data_into_db(event):
    message, channel = event.message, event.channel
    content = json.dumps(message)
    data = Flow(channel, content)
    context.db.session.add(data)
    context.db.session.commit()



from flask import Blueprint
from flask_restful import Resource, Api
from .common import Status, WorkMode, Manifest
from .flow import Flow, FlowList
from .mock import MockData, MockGroup, ActivatedMockGroup
from .config import Conf, ResetConf
from .plugin import Plugin
from .menu import Menu


api = Blueprint('api', __name__, url_prefix='/api')
api_source = Api(api)


api_source.add_resource(Status, '/status')
api_source.add_resource(Manifest, '/manifest')
api_source.add_resource(Flow, '/flow/<string:id>')
api_source.add_resource(FlowList, '/flow')
api_source.add_resource(MockGroup, '/mock', '/mock/<string:group_id>')
api_source.add_resource(MockData, '/mock/<string:group_id>/data', '/mock/<string:group_id>/data/<string:data_id>')
api_source.add_resource(ActivatedMockGroup, '/mock/activated', '/mock/<string:group_id>/<string:action>')
api_source.add_resource(Conf, '/conf/<string:plugin_name>')
api_source.add_resource(ResetConf, '/conf/<string:plugin_name>/reset')
api_source.add_resource(WorkMode, '/mode', '/mode/<string:mode>')
api_source.add_resource(Plugin, '/plugin', '/plugin/<string:plugin_name>')
api_source.add_resource(Menu, '/menu')



import lyrebird
from flask import request


class MyUI(lyrebird.PluginView):

    def index(self):
        """
        插件首页(必选)

        如果不需要jinja模板渲染，可使用send_template_file
        self.send_template_file('index.html')
        """
        return self.render_template('index.html')

    def update_data(self):
        """
        post请求        
        """
        return f'From /api/post got form name:{request.form.get("name")}'

    def get_data(self, name=''):
        """
        get请求
        """
        return f'From /api/get/ got name {name}'

    def on_data_update(self, data):
        print('Data update', data)

    def on_create(self):
        """
        插件初始化函数（必选）
        """
        # 设置模板目录（可选，设置模板文件目录。默认值templates）
        self.set_template_root('lyrebird_demo')
        # 设置静态文件目录（可选，设置静态文件目录。默认值static）
        self.set_static_root('lyrebird_demo')
        # 添加一个POST请求的接口(可选)
        self.add_url_rule('/api/post', view_func=self.update_data, methods=['POST'])
        # 添加一个Get请求的接口(可选)
        self.add_url_rule('/api/get/<string:name>', view_func=self.get_data)
        # 设置socketio处理接口(可选)
        self.on_event('from-UI', self.on_data_update, namespace='/demo')



import socket
from pathlib import Path
from colorama import Fore, Style
from .proxy_run import run
from mitmproxy.tools.dump import DumpMaster
from mitmproxy.tools.cmdline import mitmdump
from lyrebird.base_server import ThreadServer
from lyrebird import application

"""
HTTP proxy server

Default port 4272
"""

CURRENT_PATH = Path(__file__).parent
FLOW_PATH = CURRENT_PATH/'proxy_flow.py'


class LyrebirdProxyServer(ThreadServer):

    def __init__(self):
        super().__init__()
        
        conf = application.config
        self.proxy_port = str(conf.get('proxy.port', '4272')) if conf else '4272'
        '''
        --ignore_hosts:
        The ignore_hosts option allows you to specify a regex which is matched against a host:port 
        string (e.g. “example.com:443”) of a connection. Matching hosts are excluded from interception, 
        and passed on unmodified.
        
        # Ignore everything but sankuai.com, meituan.com and dianping.com:
        --ignore-hosts '^(?!.*sankuai.*)(?!.*meituan.*)(?!.*dianping.*)'
        
        According to mitmproxy docs: https://docs.mitmproxy.org/stable/howto-ignoredomains/
        '''
        self.ignore_hosts = None
        if conf.get('proxy.filters'):
            self.ignore_hosts = '^%s' % ''.join(['(?!.*%s.*)' % i for i in conf.get('proxy.filters')])
        
        self._master = None

    def run(self):
        info_msg(f'start on {get_ip()}:{self.proxy_port}', f'{Fore.CYAN} ***请在被测设备上设置代理服务器地址***')
        mitm_arguments = [
            '-s', str(FLOW_PATH),
            '-p', self.proxy_port,
            '--ssl-insecure',
            '-q'
        ]
        if self.ignore_hosts:
            mitm_arguments += ['--ignore-hosts', self.ignore_hosts]
        run(DumpMaster, mitmdump, mitm_arguments)


def get_ip():
    """
    获取本机ip
    这种方式，在有网络的时候相对准确一些

    :return: IP地址字符串

    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('google.com', 80))
    return s.getsockname()[0]


def info_msg(*msg):
    print(f'{Fore.YELLOW}mitmproxy', *msg, Style.RESET_ALL)
